# --------------------------------------------------------------------------------
#                     type YOUR NAME and your student ID here
# --------------------------------------------------------------------------------

#!/usr/bin/env python
import rospy
import math
import numpy as np
from numpy.linalg import inv
from nav_msgs.msg import Odometry
from geometry_msgs.msg import PointStamped
from copy import deepcopy
from tf.transformations import euler_from_quaternion


def get_rotation (msg):
    global roll, pitch, yaw
    orientation_q = msg.pose.pose.orientation
    orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
    (roll, pitch, yaw) = euler_from_quaternion (orientation_list)
    return yaw

def get_quaternion_from_euler(roll, pitch, yaw):

    cr = np.cos(roll * 0.5);
    sr = np.sin(roll * 0.5);
    cp = np.cos(pitch * 0.5);
    sp = np.sin(pitch * 0.5);
    cy = np.cos(yaw * 0.5);
    sy = np.sin(yaw * 0.5);

    qw = cr * cp * cy + sr * sp * sy;
    qx = sr * cp * cy - cr * sp * sy;
    qy = cr * sp * cy + sr * cp * sy;
    qz = cr * cp * sy - sr * sp * cy;

    return [qx, qy, qz, qw]


if __name__ == "__main__":


    # Initialize node
    rospy.init_node('KF', anonymous=True)

    dt = 0.1                    # Approximate time difference between measurements
    speed = 0.7                 # Linear speed of robot
    measurement_error = 0.15    # Error in observation measurements (GPS_x , GPS_y)
    errEst = 0.1                # Error due to drifting, wind, synchronization issues between the two wheels etc

    # Initialize messages that will be visualized at the end
    kf_msg = Odometry()
    gps = PointStamped()
    noisy_odom = Odometry()

    kf_msg.header.frame_id = "odom"
    gps.header.frame_id = "odom"
    noisy_odom.header.frame_id = "odom"

    # Publisher for the Kalman filter estimation
    kf_pub = rospy.Publisher('/kalman', Odometry, queue_size=1)
    # Noisy GPS publisher
    gps_pub = rospy.Publisher('/noisy_gps', PointStamped, queue_size=1)
    # Noisy Odom publisher
    odom_pub = rospy.Publisher('/noisy_odom', Odometry, queue_size=1)

    # ROS rate
    rate = rospy.Rate(1/dt) # 10hz

    # Get initial state of the robot from the odom topic
    initial_state = rospy.wait_for_message("/odom",Odometry,timeout=None)

    # Extract required data
    initial_x = initial_state.pose.pose.position.x  # x position of robot
    initial_y = initial_state.pose.pose.position.y  # y position of robot
    initial_dir = get_rotation(initial_state)       # Initial direction of robot

    # ---------------------------- TO DO ---------------------------------------------
    # Define initial state
    mu = np.matrix([ [ ??? ],
                     [ ??? ],
                     [ ??? ],
                     [1] ])
    # --------------------------------------------------------------------------------

    odom_state = deepcopy(mu) # For visualization purposes

    S = np.eye(4)

    # Control input matrix
    B = np.matrix([[0],
                   [0],
                   [1],
                   [0]])

    # Process noise matrix
    R = np.matrix([[errEst**2,0,0,0],
                   [0,errEst**2,0,0],
                   [0, 0 , 0 ,0],
                   [0, 0 , 0 ,0]])

    while not rospy.is_shutdown():

        odom_msg = rospy.wait_for_message("/odom",Odometry,timeout=None)
        # Noisy GPS measurements
        gps_x = odom_msg.pose.pose.position.x + np.random.normal(0,measurement_error)
        gps_y = odom_msg.pose.pose.position.y + np.random.normal(0,measurement_error)
        curr_dir = get_rotation(odom_msg)
        deltaDir = (curr_dir-initial_dir) + np.random.normal(0,0.01)  # This noise simulates the drift that occurs when the robot rotates
        initial_dir = deepcopy(curr_dir)

        # ---------------------------- TO DO ---------------------------------------------
        # Control input
        u = ???

        # Distance travelled in dt-> r = u * dt (Assumption--> The robot didn't rotate during dt)

        # State transition matrix
        A = np.matrix([[1,0,0, ??? ], # x & y in global
                       [0,1,0, ???],
                       [0,0,1,0],
                       [0,0,0,1]])

        # Prediction
        mu = ???

        # Covariance Matrix
        S = ???

        # Measurements / observation (gps)
        z = ???

        # State to measurement mapping matrix
        C = ???

        # Measurement error matrix
        Q = ???

        # Kalman gain
        K = ???

        # Update state
        mu = ???

        # Update covariance matrix
        S = ???

        # --------------------------------------------------------------------------------

        # Visualize noisy odom
        odom_state = np.dot(A,odom_state) + np.dot(B,u)

        # Publishes your kf_msgs to the /kalman topic that can be visualized from rviz
        kf_msg.pose.pose.position.x = mu[0,0]
        kf_msg.pose.pose.position.y = mu[1,0]
        qx,qy,qz,qw = get_quaternion_from_euler(0,0,mu[2,0])
        kf_msg.pose.pose.orientation.x = qx
        kf_msg.pose.pose.orientation.y = qy
        kf_msg.pose.pose.orientation.z = qz
        kf_msg.pose.pose.orientation.w = qw
        kf_pub.publish(kf_msg)

        # Publish noisy GPS for visualization purposes
        gps.point.x = odom_msg.pose.pose.position.x + np.random.normal(0,measurement_error) # Noisy GP
        gps.point.y = odom_msg.pose.pose.position.y + np.random.normal(0,measurement_error) # Noisy GP
        gps_pub.publish(gps)

        # Publish the estimation based only on odometry
        noisy_odom.pose.pose.position.x = odom_state[0,0]
        noisy_odom.pose.pose.position.y = odom_state[1,0]
        qx,qy,qz,qw = get_quaternion_from_euler(0,0,odom_state[2,0])
        noisy_odom.pose.pose.orientation.x = qx
        noisy_odom.pose.pose.orientation.y = qy
        noisy_odom.pose.pose.orientation.z = qz
        noisy_odom.pose.pose.orientation.w = qw
        odom_pub.publish(noisy_odom)

        rate.sleep()
